@@ -3,16 +3,11 @@ import copy
 import getpass
 import logging
 import os
-import string
 import time
 import uuid
-from weakref import finalize
-
-try:
-    import yaml
-except ImportError:
-    yaml = False
+import warnings
 
+import yaml
 import dask
 import dask.distributed
 import distributed.security
@@ -24,10 +19,17 @@ from kubernetes_asyncio.client.rest import ApiException
 from .objects import (
     make_pod_from_dict,
     make_service_from_dict,
+    make_pdb_from_dict,
     clean_pod_template,
     clean_service_template,
+    clean_pdb_template,
 )
 from .auth import ClusterAuth
+from .utils import (
+    namespace_default,
+    escape,
+    get_external_address_for_scheduler_service,
+)
 
 logger = logging.getLogger(__name__)
 
@@ -42,10 +44,20 @@ class Pod(ProcessInterface):
     Scheduler
     """
 
-    def __init__(self, cluster, core_api, pod_template, namespace, loop=None, **kwargs):
+    def __init__(
+        self,
+        cluster,
+        core_api,
+        policy_api,
+        pod_template,
+        namespace,
+        loop=None,
+        **kwargs
+    ):
         self._pod = None
         self.cluster = cluster
         self.core_api = core_api
+        self.policy_api = policy_api
         self.pod_template = copy.deepcopy(pod_template)
         self.base_labels = self.pod_template.metadata.labels
         self.namespace = namespace
@@ -164,6 +176,7 @@ class Scheduler(Pod):
                 "--idle-timeout",
                 self._idle_timeout,
             ]
+        self.pdb = None
 
     async def start(self, **kwargs):
         await super().start(**kwargs)
@@ -184,36 +197,21 @@ class Scheduler(Pod):
             namespace=self.namespace,
             port=SCHEDULER_PORT,
         )
-        if self.service.spec.type == "LoadBalancer":
-            # Wait for load balancer to be assigned
-            start = time.time()
-            while self.service.status.load_balancer.ingress is None:
-                if (
-                    self._service_wait_timeout_s > 0
-                    and time.time() > start + self._service_wait_timeout_s
-                ):
-                    raise asyncio.TimeoutError(
-                        "Timed out waiting for Load Balancer to be provisioned."
-                    )
-                self.service = await self.core_api.read_namespaced_service(
-                    self.cluster_name, self.namespace
-                )
-                await asyncio.sleep(0.2)
+        self.external_address = await get_external_address_for_scheduler_service(
+            self.core_api, self.service
+        )
 
-            [loadbalancer_ingress] = self.service.status.load_balancer.ingress
-            loadbalancer_host = loadbalancer_ingress.hostname or loadbalancer_ingress.ip
-            self.external_address = "tcp://{host}:{port}".format(
-                host=loadbalancer_host, port=SCHEDULER_PORT
-            )
-        # FIXME Set external address when using nodeport service type
-
-        # FIXME Create an optional Ingress just in case folks want to configure one
+        self.pdb = await self._create_pdb()
 
     async def close(self, **kwargs):
         if self.service:
             await self.core_api.delete_namespaced_service(
                 self.cluster_name, self.namespace
             )
+        if self.pdb:
+            await self.policy_api.delete_namespaced_pod_disruption_budget(
+                self.cluster_name, self.namespace
+            )
         await super().close(**kwargs)
 
     async def _create_service(self):
@@ -232,7 +230,38 @@ class Scheduler(Pod):
         await self.core_api.create_namespaced_service(
             self.namespace, self.service_template
         )
-        return await self.core_api.read_namespaced_service(
+        service = await self.core_api.read_namespaced_service(
+            self.cluster_name, self.namespace
+        )
+        if service.spec.type == "LoadBalancer":
+            # Wait for load balancer to be assigned
+            start = time.time()
+            while service.status.load_balancer.ingress is None:
+                if (
+                    self._service_wait_timeout_s > 0
+                    and time.time() > start + self._service_wait_timeout_s
+                ):
+                    raise asyncio.TimeoutError(
+                        "Timed out waiting for Load Balancer to be provisioned."
+                    )
+                service = await self.core_api.read_namespaced_service(
+                    self.cluster_name, self.namespace
+                )
+                await asyncio.sleep(0.2)
+        return service
+
+    async def _create_pdb(self):
+        pdb_template_dict = dask.config.get("kubernetes.scheduler-pdb-template")
+        self.pdb_template = clean_pdb_template(make_pdb_from_dict(pdb_template_dict))
+        self.pdb_template.metadata.name = self.cluster_name
+        self.pdb_template.spec.labels = copy.deepcopy(self.base_labels)
+        self.pdb_template.spec.selector.match_labels[
+            "dask.org/cluster-name"
+        ] = self.cluster_name
+        await self.policy_api.create_namespaced_pod_disruption_budget(
+            self.namespace, self.pdb_template
+        )
+        return await self.policy_api.read_namespaced_pod_disruption_budget(
             self.cluster_name, self.namespace
         )
 
@@ -268,8 +297,10 @@ class KubeCluster(SpecCluster):
 
     Parameters
     ----------
-    pod_template: kubernetes.client.V1Pod
-        A Kubernetes specification for a Pod for a dask worker.
+    pod_template: (kubernetes.client.V1Pod, dict, str)
+        A Kubernetes specification for a Pod for a dask worker. Can be either a
+        ``V1Pod``, a dict representation of a pod, or a path to a yaml file
+        containing a pod specification.
     scheduler_pod_template: kubernetes.client.V1Pod (optional)
         A Kubernetes specification for a Pod for a dask scheduler.
         Defaults to the pod_template.
@@ -317,8 +348,8 @@ class KubeCluster(SpecCluster):
     You can also create clusters with worker pod specifications as dictionaries
     or stored in YAML files
 
-    >>> cluster = KubeCluster.from_yaml('worker-template.yml')
-    >>> cluster = KubeCluster.from_dict({...})
+    >>> cluster = KubeCluster('worker-template.yml')
+    >>> cluster = KubeCluster({...})
 
     Rather than explicitly setting a number of workers you can also ask the
     cluster to allocate workers dynamically based on current workload
@@ -346,7 +377,7 @@ class KubeCluster(SpecCluster):
 
     >>> pip = 'pyarrow gcsfs git+https://github.com/dask/distributed'
     >>> conda = '-c conda-forge scikit-learn'
-    >>> KubeCluster.from_yaml(..., env={'EXTRA_PIP_PACKAGES': pip,
+    >>> KubeCluster(..., env={'EXTRA_PIP_PACKAGES': pip,
     ...                                 'EXTRA_CONDA_PACKAGES': conda})
 
     You can also start a KubeCluster with no arguments *if* the worker template
@@ -363,8 +394,6 @@ class KubeCluster(SpecCluster):
 
     See Also
     --------
-    KubeCluster.from_yaml
-    KubeCluster.from_dict
     KubeCluster.adapt
     """
 
@@ -388,17 +417,40 @@ class KubeCluster(SpecCluster):
         scheduler_pod_template=None,
         **kwargs
     ):
+        if isinstance(pod_template, str):
+            with open(pod_template) as f:
+                pod_template = dask.config.expand_environment_variables(
+                    yaml.safe_load(f)
+                )
+        if isinstance(pod_template, dict):
+            pod_template = make_pod_from_dict(pod_template)
+
         self.pod_template = pod_template
         self.scheduler_pod_template = scheduler_pod_template
-        self._generate_name = name
-        self._namespace = namespace
-        self._n_workers = n_workers
-        self._idle_timeout = idle_timeout
-        self._deploy_mode = deploy_mode
-        self._protocol = protocol
-        self._interface = interface
-        self._dashboard_address = dashboard_address
-        self._scheduler_service_wait_timeout = scheduler_service_wait_timeout
+        self._generate_name = dask.config.get("kubernetes.name", override_with=name)
+        self.namespace = dask.config.get(
+            "kubernetes.namespace", override_with=namespace
+        )
+        self._n_workers = dask.config.get(
+            "kubernetes.count.start", override_with=n_workers
+        )
+        self._idle_timeout = dask.config.get(
+            "kubernetes.idle-timeout", override_with=idle_timeout
+        )
+        self._deploy_mode = dask.config.get(
+            "kubernetes.deploy-mode", override_with=deploy_mode
+        )
+        self._protocol = dask.config.get("kubernetes.protocol", override_with=protocol)
+        self._interface = dask.config.get(
+            "kubernetes.interface", override_with=interface
+        )
+        self._dashboard_address = dask.config.get(
+            "kubernetes.dashboard_address", override_with=dashboard_address
+        )
+        self._scheduler_service_wait_timeout = dask.config.get(
+            "kubernetes.scheduler-service-wait-timeout",
+            override_with=scheduler_service_wait_timeout,
+        )
         self.security = security
         if self.security and not isinstance(
             self.security, distributed.security.Security
@@ -406,9 +458,9 @@ class KubeCluster(SpecCluster):
             raise RuntimeError(
                 "Security object is not a valid distributed.security.Security object"
             )
-        self.host = host
-        self.port = port
-        self.env = env
+        self.host = dask.config.get("kubernetes.host", override_with=host)
+        self.port = dask.config.get("kubernetes.port", override_with=port)
+        self.env = dask.config.get("kubernetes.env", override_with=env)
         self.auth = auth
         self.kwargs = kwargs
         super().__init__(**self.kwargs)
@@ -442,7 +494,7 @@ class KubeCluster(SpecCluster):
         pod_template.metadata.labels["dask.org/component"] = pod_type
         pod_template.metadata.labels["user"] = escape(getpass.getuser())
         pod_template.metadata.labels["app"] = "dask"
-        pod_template.metadata.namespace = self._namespace
+        pod_template.metadata.namespace = self.namespace
 
         if self.env:
             pod_template.spec.containers[0].env.extend(
@@ -456,36 +508,6 @@ class KubeCluster(SpecCluster):
         return pod_template
 
     async def _start(self):
-        self._generate_name = self._generate_name or dask.config.get("kubernetes.name")
-        self._namespace = self._namespace or dask.config.get("kubernetes.namespace")
-        self._idle_timeout = self._idle_timeout or dask.config.get(
-            "kubernetes.idle-timeout"
-        )
-        self._scheduler_service_wait_timeout = (
-            self._scheduler_service_wait_timeout
-            or dask.config.get("kubernetes.scheduler-service-wait-timeout")
-        )
-        self._deploy_mode = self._deploy_mode or dask.config.get(
-            "kubernetes.deploy-mode"
-        )
-
-        self._n_workers = (
-            self._n_workers
-            if self._n_workers is not None
-            else dask.config.get("kubernetes.count.start")
-        )
-        self.host = self.host or dask.config.get("kubernetes.host")
-        self.port = (
-            self.port if self.port is not None else dask.config.get("kubernetes.port")
-        )
-        self._protocol = self._protocol or dask.config.get("kubernetes.protocol")
-        self._interface = self._interface or dask.config.get("kubernetes.interface")
-        self._dashboard_address = self._dashboard_address or dask.config.get(
-            "kubernetes.dashboard_address"
-        )
-        self.env = (
-            self.env if self.env is not None else dask.config.get("kubernetes.env")
-        )
 
         self.pod_template = self._get_pod_template(self.pod_template, pod_type="worker")
         self.scheduler_pod_template = self._get_pod_template(
@@ -512,9 +534,10 @@ class KubeCluster(SpecCluster):
         await ClusterAuth.load_first(self.auth)
 
         self.core_api = kubernetes.client.CoreV1Api()
+        self.policy_api = kubernetes.client.PolicyV1beta1Api()
 
-        if self._namespace is None:
-            self._namespace = _namespace_default()
+        if self.namespace is None:
+            self.namespace = namespace_default()
 
         self._generate_name = self._generate_name.format(
             user=getpass.getuser(), uuid=str(uuid.uuid4())[:10], **os.environ
@@ -528,14 +551,11 @@ class KubeCluster(SpecCluster):
             self.scheduler_pod_template, pod_type="scheduler"
         )
 
-        finalize(
-            self, _cleanup_resources, self._namespace, self.pod_template.metadata.labels
-        )
-
         common_options = {
             "cluster": self,
             "core_api": self.core_api,
-            "namespace": self._namespace,
+            "policy_api": self.policy_api,
+            "namespace": self.namespace,
             "loop": self.loop,
         }
 
@@ -570,12 +590,16 @@ class KubeCluster(SpecCluster):
         }
         self.worker_spec = {i: self.new_spec for i in range(self._n_workers)}
 
+        self.name = self.pod_template.metadata.generate_name
+
         await super()._start()
 
     @classmethod
     def from_dict(cls, pod_spec, **kwargs):
         """Create cluster with worker pod spec defined by Python dictionary
 
+        Deprecated, please use the `KubeCluster` constructor directly.
+
         Examples
         --------
         >>> spec = {
@@ -598,12 +622,17 @@ class KubeCluster(SpecCluster):
         --------
         KubeCluster.from_yaml
         """
-        return cls(make_pod_from_dict(pod_spec), **kwargs)
+        warnings.warn(
+            "KubeCluster.from_dict is deprecated, use the constructor directly"
+        )
+        return cls(pod_spec, **kwargs)
 
     @classmethod
     def from_yaml(cls, yaml_path, **kwargs):
         """Create cluster with worker pod spec defined by a YAML file
 
+        Deprecated, please use the `KubeCluster` constructor directly.
+
         We can start a cluster with pods defined in an accompanying YAML file
         like the following:
 
@@ -629,22 +658,10 @@ class KubeCluster(SpecCluster):
         --------
         KubeCluster.from_dict
         """
-        if not yaml:
-            raise ImportError(
-                "PyYaml is required to use yaml functionality, please install it!"
-            )
-        with open(yaml_path) as f:
-            d = yaml.safe_load(f)
-            d = dask.config.expand_environment_variables(d)
-            return cls.from_dict(d, **kwargs)
-
-    @property
-    def namespace(self):
-        return self.pod_template.metadata.namespace
-
-    @property
-    def name(self):
-        return self.pod_template.metadata.generate_name
+        warnings.warn(
+            "KubeCluster.from_yaml is deprecated, use the constructor directly"
+        )
+        return cls(yaml_path, **kwargs)
 
     def scale(self, n):
         # A shim to maintain backward compatibility
@@ -685,61 +702,3 @@ class KubeCluster(SpecCluster):
                 logs[key] = log
 
         return logs
-
-
-def _cleanup_resources(namespace, labels):
-    """ Remove all pods with these labels in this namespace """
-    import kubernetes
-
-    core_api = kubernetes.client.CoreV1Api()
-
-    pods = core_api.list_namespaced_pod(namespace, label_selector=format_labels(labels))
-    for pod in pods.items:
-        try:
-            core_api.delete_namespaced_pod(pod.metadata.name, namespace)
-            logger.info("Deleted pod: %s", pod.metadata.name)
-        except kubernetes.client.rest.ApiException as e:
-            # ignore error if pod is already removed
-            if e.status != 404:
-                raise
-
-    services = core_api.list_namespaced_service(
-        namespace, label_selector=format_labels(labels)
-    )
-    for service in services.items:
-        try:
-            core_api.delete_namespaced_service(service.metadata.name, namespace)
-            logger.info("Deleted service: %s", service.metadata.name)
-        except kubernetes.client.rest.ApiException as e:
-            # ignore error if service is already removed
-            if e.status != 404:
-                raise
-
-
-def format_labels(labels):
-    """ Convert a dictionary of labels into a comma separated string """
-    if labels:
-        return ",".join(["{}={}".format(k, v) for k, v in labels.items()])
-    else:
-        return ""
-
-
-def _namespace_default():
-    """
-    Get current namespace if running in a k8s cluster
-
-    If not in a k8s cluster with service accounts enabled, default to
-    'default'
-
-    Taken from https://github.com/jupyterhub/kubespawner/blob/master/kubespawner/spawner.py#L125
-    """
-    ns_path = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
-    if os.path.exists(ns_path):
-        with open(ns_path) as f:
-            return f.read().strip()
-    return "default"
-
-
-def escape(s):
-    valid_characters = string.ascii_letters + string.digits + "-"
-    return "".join(c for c in s if c in valid_characters)
